import shutil
import subprocess
from dataclasses import dataclass
from enum import Enum
from glob import glob
from pathlib import Path
from typing import Dict, Union, List

import orjson
import toml
from loguru import logger
from pbt.poetry import Poetry


class PackageType(str, Enum):
    Poetry = "poetry"


@dataclass
class Package:
    name: str
    type: PackageType
    dir: Path
    version: str
    dependencies: Dict[str, str]
    # list of packages that this package uses
    inter_dependencies: List["Package"]
    # list of packages that use the current package
    invert_inter_dependencies: List["Package"]

    def build(self):
        """Build the package"""
        outfile = f"{self.name}-{self.version}.tar.gz"
        rebuild = True
        if (self.dir / "dist" / outfile).exists():
            if not self.is_modified():
                rebuild = False

        if rebuild:
            logger.info("Build package {}", self.name)
            shutil.rmtree(str(self.dir / "dist"))
            subprocess.check_call(["poetry", "build"], cwd=str(self.dir))

    def is_modified(self) -> bool:
        """Whether the package has been modified since the last version"""
        return True

    def install_into(self, package: "Package", build: bool = False):
        """Install the current package (self) in the virtual environment of another package (`package`)"""
        if build:
            self.build()
        logger.info(
            "Install package {} into package {} environment", self.name, package.name
        )
        pipfile = package.pkg_handler.get_pip_file(package.dir)
        subprocess.check_output([pipfile, "uninstall", "-y", self.name])
        subprocess.check_output([pipfile, "install", self.get_wheel_file()])

    def update_package_version(self, package: "Package"):
        """Update the version of another package in this package"""
        assert package.name in self.dependencies
        self.pkg_handler.update_package(self, package.name, package.version)

    @property
    def pkg_handler(self):
        if self.type == PackageType.Poetry:
            return Poetry
        raise NotImplementedError(self.type)

    def get_wheel_file(self):
        whl_files = glob(str(self.dir / f"dist/{self.name.replace('-', '_')}*.whl"))
        if len(whl_files) == 0:
            return None
        return whl_files[0]

    def to_dict(self):
        return {
            "name": self.name,
            "type": self.type,
            "dir": self.dir,
            "version": self.version,
            "dependencies": self.dependencies,
            "inter_dependencies": [p.name for p in self.inter_dependencies],
            "invert_inter_dependencies": [
                p.name for p in self.invert_inter_dependencies
            ],
        }

    @staticmethod
    def save(packages: Dict[str, "Package"], outfile: Union[str, Path]):
        with open(str(outfile), "wb") as f:
            f.write(
                orjson.dumps(
                    {package.name: package.to_dict() for package in packages.values()}
                )
            )

    @staticmethod
    def load(infile: str) -> Dict[str, "Package"]:
        with open(infile, "r") as f:
            raw_packages = orjson.loads(f.read())
        packages = {name: Package(**o) for name, o in raw_packages}
        for package in packages:
            package.type = PackageType(package.type)
            packages.inter_dependencies = [
                packages[name] for name in package.inter_dependencies
            ]
            packages.invert_inter_dependencies = [
                packages[name] for name in package.invert_inter_dependencies
            ]
        return packages


def search_packages(root_dir: Path):
    logger.info("Search packages...")
    packages = {}

    for poetry_file in glob(str(root_dir / "*/pyproject.toml")):
        poetry_file = Path(poetry_file)
        package_dir = poetry_file.parent

        try:
            with open(poetry_file, "r") as f:
                project_cfg = toml.loads(f.read())
                package_name = project_cfg["tool"]["poetry"]["name"]
                package_version = project_cfg["tool"]["poetry"]["version"]
                package_dependencies = project_cfg["tool"]["poetry"]["dependencies"]
        except:
            logger.error("Error while parsing configuration in {}", package_dir)
            raise

        logger.info("Found package {}", package_name)
        packages[package_name] = Package(
            name=package_name,
            type=PackageType.Poetry,
            version=package_version,
            dependencies=package_dependencies,
            dir=package_dir,
            inter_dependencies=[],
            invert_inter_dependencies=[],
        )

    package_names = set(packages.keys())
    for package in packages.values():
        for pname in package_names.intersection(package.dependencies.keys()):
            package.inter_dependencies.append(packages[pname])
            packages[pname].invert_inter_dependencies.append(package)

    return packages



import os
import subprocess
from pathlib import Path

import click
from typing import List

from loguru import logger

from pbt.package import search_packages
from pbt.pypi import PyPI

ROOT_DIR = Path(os.path.abspath(__file__)).parent.parent.parent
CACHE_DIR = ROOT_DIR / ".cache"


@click.group()
def cli():
    pass


@click.command()
@click.option(
    "-p",
    "--package",
    multiple=True,
    help="Specify the active package that we are working on. If empty, use all packages",
)
def make(package: List[str]):
    active_packages = package
    packages = search_packages(ROOT_DIR)
    if len(active_packages) == 0:
        active_packages = packages.keys()
    active_packages = set(active_packages)

    for package in packages.values():
        if package.is_modified():
            logger.info("Package {} has been modified", package.name)
            parent_packages = [
                pp
                for pp in package.invert_inter_dependencies
                if pp.name in active_packages
            ]
            if len(parent_packages) > 0:
                package.build()
                for pp in parent_packages:
                    package.install_into(pp, build=False)
                    pp.update_package_version(package)


@click.command()
def publish():
    packages = search_packages(ROOT_DIR)
    for package in packages.values():
        if PyPI.does_package_exist(package.name, package.version):
            logger.info("Skip publishing package {} as it exists", package.name)
            continue
        logger.info("Publish package {}", package.name)
        package.build()
        for pp in package.invert_inter_dependencies:
            pp.update_package_version(package)
        subprocess.check_call(["poetry", "publish"], cwd=str(package.dir))


cli.add_command(make)
cli.add_command(publish)


if __name__ == "__main__":
    cli()



from typing import Optional
import requests


class PyPI:
    index = "https://pypi.org"

    @classmethod
    def set_test_package_index(cls):
        cls.index = "https://test.pypi.org"

    @classmethod
    def does_package_exist(
        cls, package_name: str, package_version: Optional[str] = None
    ) -> bool:
        """Check if a package exist in pypi index"""
        resp = requests.get(cls.index + f"/pypi/{package_name}/json")
        if resp.status_code == 404:
            return False

        assert resp.status_code == 200
        if package_version is None:
            return True
        releases = resp.json()['releases']
        return package_version in releases



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pbt']

package_data = \
{'': ['*']}

install_requires = \
['orjson>=3.6.3,<4.0.0', 'toml>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'polyrepo-bt',
    'version': '0.1.0',
    'description': 'A build tool for poetry packages (living in submodules of a Git repo)',
    'long_description': None,
    'author': 'Binh Vu',
    'author_email': 'binh@toan2.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os
import re
import subprocess
import sys
from pathlib import Path
from typing import Union, TYPE_CHECKING

from loguru import logger

if TYPE_CHECKING:
    from pbt.package import Package


class Poetry:
    @classmethod
    def get_python_file(cls, package_dir: Union[str, Path]):
        if sys.prefix != sys.base_prefix:
            logger.error(
                "Can't obtain the python executable file because the virtual environment is active."
            )
            raise Exception(
                "Can't obtain the python executable file because the virtual environment is active."
            )

        pres = subprocess.run(
            ["poetry", "env", "info", "-p"],
            cwd=str(package_dir),
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
        )
        python_file = pres.stdout.decode().strip()
        assert "\n" not in python_file
        python_file = Path(python_file) / "bin/python"
        assert python_file.exists()
        return python_file

    @classmethod
    def get_pip_file(cls, package_dir: Union[str, Path]):
        return Poetry.get_python_file(package_dir).parent / "pip"

    @classmethod
    def update_package(
        cls, target_package: "Package", package_name: str, package_version: str
    ):
        should_update = False
        with open(os.path.join(target_package.dir, "pyproject.toml"), "r") as f:
            lines = f.readlines()
            match_lines = [
                (i, line)
                for i, line in enumerate(lines)
                if re.match(f"{package_name} *= *", line) is not None
            ]
            assert len(match_lines) == 1
            idx, line = match_lines[0]
            m = re.match(
                rf"""({package_name} *= *)(?:'|")([^0-9]*)([^'"]+)(?:'|")(.*)""", line,
                flags=re.DOTALL
            )
            if m is not None:
                groups = m.groups()
                prev_version = groups[2]
                if prev_version != package_version:
                    should_update = True
                    logger.info(
                        f"In {target_package.name}, bump {package_name} from `{prev_version}` to `{package_version}`"
                    )
                    lines[idx] = f'{groups[0]}"{groups[1]}{package_version}"{groups[3]}'
            else:
                raise NotImplementedError(f"Do not know how to parse `{line}` yet")

        if should_update:
            with open(os.path.join(target_package.dir, "pyproject.toml"), "w") as f:
                for line in lines:
                    f.write(line)
